//=================================================================================
//=================================================================================
///
/// \file	 ImageProcessingDemo.cpp
///
/// \brief
///			 
///	\par	 Contents
///				- Shrink image
///				- Gauss filtering
///				- Histogram stretch
///				- Gradient / energy calculation
///				- Color image representation (RGB/HSV)
///				- Histogram with bins
///				- Color segmentation
///
/// \author  Marc Schlipsing
///
/// \date    2010-04-06
///
/// \par     Copyright (c) 2010:
///          Institut fr Neuroinformatik
///          Universitt Bochum
//=================================================================================
//=================================================================================


#include "PgmIO.h"
#include "PpmIO.h"


static void filterGauss3x1( unsigned char * pImgDst, const unsigned char * pImgSrc, const int width, const int height )
{
	for (int y = 0; y < height; ++y)
	{
		unsigned char * pDst = pImgDst + y * width + 1;
		const unsigned char * pSrc = pImgSrc + y * width;

		for (int x = 1; x < width-1; ++x)
		{
			const unsigned int sum = (int)pSrc[0] + 2*(int)pSrc[1] + (int)pSrc[2];
			*pDst = sum / 4;
			++pDst;
			++pSrc;
		}
	}
};

static void filterGauss1x3( unsigned char * pImgDst, const unsigned char * pImgSrc, const int width, const int height )
{
	for (int y = 1; y < height-1; ++y)
	{
		unsigned char * pDst = pImgDst + y * width;
		const unsigned char * pSrc = pImgSrc + y * width;

		for (int x = 0; x < width; ++x)
		{
			const unsigned int sum = (int)pSrc[-width] + 2*(int)pSrc[0] + (int)pSrc[width];
			*pDst = sum / 4;
			++pDst;
			++pSrc;
		}
	}
};

//////////////////////////////////////////////////////////////////////////
// Filter image with 3x3 Gaussian kernel (separated in 3x1 and 1x3)
//////////////////////////////////////////////////////////////////////////
static void filterGaussian3x3( unsigned char * pImg, const int width, const int height )	// 1 2 1
{																							// 2 4 2
	unsigned char * pFltY = new unsigned char[ width * height ];							// 1 2 1

	filterGauss1x3( pFltY, pImg, width, height );

	// copy border
	memcpy( pFltY, pImg, width );
	memcpy( pFltY+width*(height-1), pImg+width*(height-1), width );

	filterGauss3x1( pImg, pFltY, width, height );		

	delete pFltY;
};


int main( int argc, char* argv[] )
{
	//////////////////////////////////////////////////////////////////////////
	// Read gray-value image
	//////////////////////////////////////////////////////////////////////////
	unsigned char * pImage = 0;
	int width, height;

	bool readOk = readPGM( "../eyes_dark.pgm", &pImage, width, height );

	if( ! readOk )
	{
		printf( "Reading image failed!\n");
		printf( "Press any key to exit.\n" );
		getchar();
		return -1;
	}


	//////////////////////////////////////////////////////////////////////////
	// Scale image by 1/2
	//////////////////////////////////////////////////////////////////////////
	const int widthScl = width / 2;
	const int heightScl = height / 2;

	unsigned char * pScaledImage = new unsigned char[ widthScl * heightScl ];

	for( int y = 0; y < heightScl; y++ )
	{
		for( int x = 0; x < widthScl; x++ )
		{
			pScaledImage[ x + y*widthScl ] = pImage[ 2*x + 2*y*width ];
		}
	}

	writePGM( "../half.pgm", pScaledImage, widthScl, heightScl );


	//////////////////////////////////////////////////////////////////////////
	// Filter image with 3x3 Gaussian kernel
	//////////////////////////////////////////////////////////////////////////
	filterGaussian3x3( pScaledImage, widthScl, heightScl );
	
	writePGM( "../halfFiltered.pgm", pScaledImage, widthScl, heightScl );


	//////////////////////////////////////////////////////////////////////////
	// Compute histogram / cut upper and lower 5% of gray-values
	//////////////////////////////////////////////////////////////////////////
	unsigned int histogram[256];
	memset( histogram, 0, 256*sizeof(unsigned int) );

	// calculate histogram
	for( int i = 0; i < widthScl*heightScl; i++ )
		++histogram[ pScaledImage[i] ];

	// determine lower and upper bound for histogram stretch
	const float			cutOffPercentage = 0.05;
	unsigned char		lowerBound, upperBound;
	unsigned int		histAccu = 0;
	const unsigned int	lowerPercentile = cutOffPercentage * widthScl*heightScl;
	const unsigned int	upperPercentile = (1-cutOffPercentage) * widthScl*heightScl;

	for( int h = 0; h < 256 ; h++ )
	{
		histAccu += histogram[h];
		if( histAccu <= lowerPercentile )
		{
			lowerBound = h;
			continue;
		}
		if( histAccu >= upperPercentile )
		{
			upperBound = h;
			break;
		}
	}

	// assign new gray-values from linear mapping between lower and upper bound
	const float histScale = 255. / (upperBound - lowerBound);
	for( int i = 0; i < widthScl*heightScl; i++ )
	{
		const int newVal = histScale * ( (int)pScaledImage[i] - lowerBound );
		pScaledImage[i] = std::min<int>( 255, std::max<int>( 0, newVal ) );
	}

	writePGM( "../histogram.pgm", pScaledImage, widthScl, heightScl );


	//////////////////////////////////////////////////////////////////////////
	// Compute image energy from gradients
	//////////////////////////////////////////////////////////////////////////
	unsigned char * energy = new unsigned char[ widthScl * heightScl ];
	memset( energy, 0, widthScl*heightScl );

	for (int y = 1; y < heightScl-1; ++y)
	{
		const int rowOffset = y * widthScl;
		for (int x = 1; x < widthScl-1; ++x)
		{
			const int gradX = pScaledImage[rowOffset+x+1] - pScaledImage[rowOffset+x-1];
			const int gradY = pScaledImage[rowOffset+x+widthScl] - pScaledImage[rowOffset+x-widthScl];
			energy[rowOffset+x] = sqrt( (float)(gradX * gradX + gradY * gradY) ) / sqrt(2.f);
		}
	}
	
	writePGM( "../energy.pgm", energy, widthScl, heightScl );


	//////////////////////////////////////////////////////////////////////////
	// Segment high energy areas by Thresholding
	//////////////////////////////////////////////////////////////////////////
	for( int i = 0; i < widthScl*heightScl; i++ )
	{
		if( energy[i] > 30 )
			energy[i] = 255;
		else
			energy[i] = 0;
	}
	writePGM( "../energyThresh.pgm", energy, widthScl, heightScl );

	delete energy;
	delete pScaledImage;
	free( pImage );



	//////////////////////////////////////////////////////////////////////////
	// Read color (RGB) image
	//////////////////////////////////////////////////////////////////////////
	rtcvRgbaValue * pRgbImage = 0;
	//readOk = readPPM( "../eyes_color.ppm", &pRgbImage, width, height );
	readOk = readPPM( "../HSV_cone.ppm", &pRgbImage, width, height ); // proof of concept ;)

	if( ! readOk )
	{
		printf( "Reading color image failed!\n");
		printf( "Press any key to exit.\n" );
		getchar();
		return -1;
	}


	//////////////////////////////////////////////////////////////////////////
	// Determine the dominant color from a Hue-histogram
	//////////////////////////////////////////////////////////////////////////

	// Compute hue-histogram with 8 bins
	unsigned int hueHist[8] = {0,0,0,0,0,0,0,0};
	for( int i = 0; i < width*height; i++ )
		++hueHist[ pRgbImage[i].getHue() >> 5 ];

	// Find maximum in histogram
	int maxHist = 0;
	int maxHue = -1;
	for( int h = 0; h < 8; h++ )
	{
		if( hueHist[h] > maxHist )
		{
			maxHist = hueHist[h];
			maxHue = h;
		}
	}

	//////////////////////////////////////////////////////////////////////////
	// Segment dominant color (and neighbors) in HSV color space
	//////////////////////////////////////////////////////////////////////////
	unsigned char * hueSeg = new unsigned char[ width*height ];
	for( int i = 0; i < width*height; i++ )
	{
		const unsigned char hue = pRgbImage[i].getHue() >> 5; // in bins
		const unsigned char sat = pRgbImage[i].getSat();
		const unsigned char val = pRgbImage[i].getV();
		if( hue == maxHue && sat > 100 && val > 100 )
			hueSeg[i] = 255;
		else
			hueSeg[i] = 0;
	}

	writePGM( "../hueSegmentation.pgm", hueSeg, width, height );

	delete hueSeg;
	free( pRgbImage );
	
	printf( "Finished! Press any key.\n" );
	getchar();

	return 0;
}